#!/usr/bin/perl

package main;
################################################################################
#       CONFIGURATION SECTION                                                  #
################################################################################

################ TMPRED OPTIONS ################
###### The TMPRED_BIN constant is the location of the tmpred executable
use constant TMPRED_BIN => '../lib/tmpred/tmpred';
###### The TMPRED_DAT constant is the location of the matrix.tab data file 
###### downloaded with tmpred
use constant TMPRED_DAT => '../lib/tmpred/matrix.tab';

################ PSORT OPTIONS ################ 
###### The PSORT_BIN constant is the location of the psort executable
use constant PSORT_BIN => '../lib/psort/psort';

################ TEMP OPTIONS ################ 
###### The JOB_TEMPLATE constant is the location of a writable directory to 
###### store temporary files. The parent directory must exists and every 'X' is
###### replaced with a random letter. For example, if the template is 
###### '/tmp/job.XXXX', then /tmp must exist and be writable, and job.XXXX will
###### become something like job.A4FM
use constant JOB_TEMPLATE => '../tmp/job.XXXX';

################ PHOSPHORYLATION OPTIONS
###### List of amino acids consided phosphorylated with modified weights
use constant PHOSPHORYLATION_MODS => ('Y +79.9663', 'T +79.9663', 'S +79.9663' );

################################################################################
#       END CONFIGURATION SECTION                                              #
################################################################################

my $HELP = <<HELP;
DDF2GO is used to translate high throughput DDF proteomics data into higher 
quality, experiment-inferred, GO CC annotations. For more information, see paper
"A high-throughput experiment-based method to identify subcellular localization" 
Lakshmi R Pillai, et al.

MANDATORY PARAMETERS:
    -i, --input_files FILE...       A list of files with protein/peptide 
                                        information in the format of PepMOD 
                                        output

DATABASE PARAMETERS:
    -d, --fasta FILE                FASTA database file containing the proteins 
                                        used in search. The headers must match 
                                        the input file protein names. If a FASTA
                                        database is not given, DDF2GO will try to 
                                        parse GI numbers from the protein names 
                                        and download the sequences from NCBI.
                                        (Skipping is NOT recommended)
    --save_fasta FILE               If sequences are downloaded from NCBI, this 
                                        option save them in a file for later use.

OUTPUT PARAMETERS:
    -p, --map, --mapped_file FILE   Output file where all the clearly mapped GO 
                                        CC terms are written. (Default: STDOUT)
    -r, --mar, --margin_file FILE   Output file where all the marginal DDF 
                                        Profiles/TMD are written. A DDF 
                                        Profile/TMD is marginal if it is within 
                                        2 percentage points of a category boundary 
                                        (Default: STDERR)

MISCELLANEOUS PARAMETERS:
    --frac REGEX REGEX REGEX REGEX  4 regular expressions that are used to map the 
                                        sample names to the DDF fractions 
                                        (Default: 1 2 3 4)
    --id_mod                        Regular expression used to truncate the 
                                        protein ids to match the fasta headers. 
                                        The sub-string matched by the regular 
                                        expression is removed.
    --debug                         Keeps the program from deleting the temporary
                                        files on exit
    --phosphorylation               Marks DDF Fraction that contain phosphorylated
                                        amino acids 
                                        (Y +79.9663, T +79.9663, S +79.9663)
    --configure                     Reconfigure DDF2GO

HELP


use v5.10.1;
use warnings;
use strict;
use threads;

use Data::Dumper;
#use Parse::Fasta;
use File::Temp;
use File::Spec::Functions qw/catfile/;
use File::Basename qw/fileparse/;
use File::Path qw/remove_tree/;
use LWP::Simple;
use LWP::UserAgent;
use HTTP::Request::Common;
use Getopt::Long;
use IO::File;
use IO::Handle;
use List::Util qw/sum/;
use Cwd;

use constant EPOST   => "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/epost.fcgi";
use constant EFETCH  => "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi";
use constant ESEARCH => "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi";


use constant RANDOM_CHAR => (split //, 'QWERTYUIOPASDFGHJKLZXCVBNMqqwertyuiopasdfghjklzxcvbnm1234567890_');
use constant RANDOM_LENGTH => scalar @{[(RANDOM_CHAR)]};



local $\ = $/;
local $, = "\t";

my $arguments = {debug => 0, files => [], frac => []};

die $HELP unless GetOptions (   "frac=s{4}"           => $arguments->{frac},
                                "input_files|i=s{,}"  => $arguments->{files},
                                "debug"               => \$arguments->{debug},
                                "fasta|d=s"           => \$arguments->{fasta},
                                "save_fasta=s"        => \$arguments->{save},
                                "mapped_file|map|p=s" => \$arguments->{output},
                                "margin_file|mar|r=s" => \$arguments->{margin},
                                "phosphorylation"     => \$arguments->{phosphorylation},
                                "id_mod=s"            => \$arguments->{id_mod},
                                "help"                => sub{die $HELP; exit;},
                                "configure"           => sub{configure(); exit;},
                            );

unless(TMPRED_BIN && TMPRED_DAT && PSORT_BIN && JOB_TEMPLATE){
    configure();
    exit;
}

$arguments->{output} = new IO::File ($arguments->{output}, 'w') || IO::Handle->new_from_fd(fileno(STDOUT), 'w');
$arguments->{margin} = new IO::File ($arguments->{margin}, 'w') || IO::Handle->new_from_fd(fileno(STDERR), 'w');
$arguments->{frac} = [1..4] unless (@{$arguments->{frac}});
                            
push @{$arguments->{files}}, @ARGV;

die $HELP unless(@{$arguments->{files}});

$File::Temp::KEEP_ALL = $arguments->{debug};

(my $job_dir = JOB_TEMPLATE) =~ s/X/ (RANDOM_CHAR)[int(rand(RANDOM_LENGTH))] /eg;
mkdir $job_dir;

my $data = {};
my $fractions = {};

foreach my $file ( @{$arguments->{files}} ) {
    open FILE, $file;

    while(<FILE>){
        last if /^Description/;
    };

    while(<FILE>){
        s/\r?\n//;
        next if (/^\d*\s/ || /^Scan/);
        
        my $d = [split "\t"];

        my $acc = $d->[0];
        $acc =~ s/$arguments->{id_mod}// if $arguments->{id_mod};
        
        my $headers = {};

        for(my $tell = tell(FILE); defined($_ = <FILE>); $tell = tell(FILE)){
            unless (/^\d/ || /^Scan/){
                seek(FILE, $tell, 0);
                last;
            }
            if( /^Scan/ ){
                my $c = 0;
                $headers = {map {$_ => $c++} split "\t"};
                next;
            }

            s/\r?\n//;
            
            my $d = [split "\t"];
            

            foreach my $frac (@{$arguments->{frac}}){
                next unless ($d->[-1] =~ /$frac/);
            
                $data->{$acc}->{frac}->{$frac} += $d->[$headers->{HyperScore}];  

                $data->{$acc}->{phos}->{$frac} ||= (grep {index($d->[$headers->{Modifications}], $_) >= 0} (PHOSPHORYLATION_MODS));
                last;
            }
        }
    }

    close FILE;
}

foreach my $v (values %$data){
    my $sum = sum grep {$_} @{$v->{frac}}{@{$arguments->{frac}}};
    foreach my $frac (@{$arguments->{frac}}){
        $v->{frac}->{$frac} ||= 0;
        $v->{frac}->{$frac} /= $sum;
        $v->{frac}->{$frac} *= 100;
        $v->{frac}->{$frac} = sprintf("%.0f", $v->{frac}->{$frac});
    }
}

$arguments->{fasta} = new IO::File ($arguments->{fasta}) || 
                      get_fasta_database( (new IO::File ($arguments->{save}, '+>') || 
                                           new File::Temp ( TEMPLATE => catfile($job_dir, "fasta.db.XXXX")) ), 
                                         map { /gi\|(\d+)/; $1 } keys %$data);

foreach my $fh (@$arguments{ qw/ output margin /}){
print $fh join("\t", "Gi number",
                 "Protein",
                 "D|T|TD|S",
                 "Average TMD",
                 (($fh == $arguments->{margin})?'':"GO CC")
            );
}

my $fasta = new Parse::Fasta $arguments->{fasta};
open(LINK, '>', catfile($job_dir, 'link.txt'));
while(my ($name, $sequence) = $fasta->next){

    my $temp_dir;
    my $id = [ grep { cmp_id($name, $_) } keys %$data ];
    
    if(@$id == 1){
        $id = $id->[0];
    }elsif(@$id > 1){
        warn "$name has multiple matches : " . map { "\t$_\n" } @$id;
        next;
    }else{
        next;
    }

    do { $temp_dir = catfile($job_dir, join('', map { (RANDOM_CHAR)[int(rand(RANDOM_LENGTH))] } (1..6))) } until(mkdir $temp_dir);
    print LINK $id, $name, $temp_dir;
    
    my $tmpread_thr = threads->create({'context' => 'list'}, \&tmpred_prediction, ($sequence, $temp_dir));
    my $psort_thr   = threads->create(\&psort_prediction, ($sequence, $temp_dir));

    $data->{$id}->{name} = $name;

    @{$data->{$id}->{tmpred}}{ qw/ type tmd/ } = $tmpread_thr->join();
    $data->{$id}->{psort}->{tmd} = $psort_thr->join();

    $data->{$id}->{tmd} = (($data->{$id}->{tmpred}->{tmd} + $data->{$id}->{psort}->{tmd}) / 2);

    $data->{$id}->{go} = go_mapping([@{$data->{$id}->{frac}}{@{$arguments->{frac}}}], $data->{$id}->{tmd});

    my $frac = {%{$data->{$id}->{frac}}};
    foreach (keys %$frac){
        $frac->{$_} = '*' . $frac->{$_} if($arguments->{phosphorylation} && $data->{$id}->{phos}->{$_});
    }
    
    my $fh = $arguments->{output};
    foreach(@{$data->{$id}->{frac}}{@{$arguments->{frac}}}){
        if(($_ >= 8 && $_ <=12) || ($_ >= 88 && $_ <=92)){
            $fh = $arguments->{margin};
        }
    }
    my $T = $data->{$id}->{frac}->{$arguments->{frac}->[1]};
    $fh = $arguments->{margin} if($T >= 68 && $T <= 72);
    
    $data->{$id}->{go} = [] if( $fh == $arguments->{margin} );
    
    print $fh $id, $name, join('|', @$frac{@{$arguments->{frac}}}), $data->{$id}->{tmd}, join('|', @{$data->{$id}->{go}});

    delete $data->{$id};

    remove_tree($temp_dir) unless $arguments->{debug};
}
close LINK;
remove_tree($job_dir) unless $arguments->{debug};

die "The following IDs were not found in the fasta database:\n",
        join("\n", map {"\t$_"} keys %$data) if(keys %$data);

exit;
#print simple_output $fractions, $data;

sub cmp_id {
    my ($a, $b) = sort {length($b) <=> length($a)} @_ ;

    return (index($a, $b)+1);
}

sub tmpred_prediction{
    $SIG{'KILL'} = sub { threads->exit(); };

    my $sequence = shift;
    my $temp_dir = shift;
    mkdir catfile($temp_dir, 'tmpred');

    my $tmp_seq  = new File::Temp TEMPLATE => catfile($temp_dir, 'tmpred', "sequence.XXXX");
    my $tmp_out1 = new File::Temp TEMPLATE => catfile($temp_dir, 'tmpred', "output.1.XXXX");
    my $tmp_out2 = new File::Temp TEMPLATE => catfile($temp_dir, 'tmpred', "output.2.XXXX");
    my $tmp_out3 = new File::Temp TEMPLATE => catfile($temp_dir, 'tmpred', "output.3.XXXX");
    my $tmp_err  = new File::Temp TEMPLATE => catfile($temp_dir, 'tmpred', "error.XXXX");

    print $tmp_seq join("\n", grep {$_} split /(.{60})/, $sequence);
    close $tmp_seq;

#    print STDERR TMPRED_BIN . " -par=" . TMPRED_DAT . " -min=17 -max=33 -def -in=" . $tmp_seq->filename .
#                                " -out=". $tmp_out1->filename ." -out2=". $tmp_out2->filename .
#                                " -out3=". $tmp_out3->filename;

    open(IN, '-|', TMPRED_BIN . " -par=" . TMPRED_DAT . " -min=17 -max=33 -def -in=" . $tmp_seq->filename .
                                " -out=". $tmp_out1->filename ." -out2=". $tmp_out2->filename .
                                " -out3=". $tmp_out3->filename);
    print $tmp_err do{ local $/; <IN> }; 
    close IN;
    
    my $output = do { local $/; <$tmp_out1> };

    if($output =~ m/-----> (\S+) prefered model: .*?\s*(\d+) strong transmembrane helices/){

        return $1, $2;

    }else{
        return "NONE", 0;
    }
}

sub psort_prediction{
    $SIG{'KILL'} = sub { threads->detach(); threads->exit();};

    my $sequence = shift;
    my $temp_dir = shift;
    mkdir catfile($temp_dir, 'psort');

    my $tmp_seq  = new File::Temp TEMPLATE => catfile($temp_dir, 'psort', "sequence.XXXX");

    print $tmp_seq ">header";
    print $tmp_seq join("\n", grep {$_} split /(.{60})/, $sequence);
    close $tmp_seq;

    open(PSORT, '-|', PSORT_BIN, $tmp_seq->filename);
    my $output = do { local $/; <PSORT> };
    close PSORT;

    my $info = { map { split /: / } grep { /: / } split /\s{2}/, $output };

    $info->{tms} /= 0.08; # The simple output for psort has each tms equal to aprox. 0.08
    $info->{tms} = 0 if($info->{tms} < 0);

    if($arguments->{debug}){
        my $tmp_out_long  = new File::Temp TEMPLATE => catfile($temp_dir, 'psort', "output.verb.XXXX");
        my $tmp_out_short = new File::Temp TEMPLATE => catfile($temp_dir, 'psort', "output.simp.XXXX");

        print $tmp_out_short $output;
        
        open IN, '-|', PSORT_BIN, '-v', $tmp_seq->filename;
        print $tmp_out_long do{ local $/; <IN> };
        close IN;
    }

    return int $info->{tms};
}

sub go_mapping {
    my $frac = shift;
    my $tmd  = shift;
    my $return = [];
    
    foreach my $f (@$frac){
        $f = 0 if($f < 10);
        if($f >= 90){
            $_ = 0 foreach (@$frac);
            $f = 100;
            last;
        }
    }
    
    my $name = {};
    @$name{ qw/D T TD S/ } = @$frac;
    $name = join '/', grep {$name->{$_}} qw/D T TD S/;
    
#    print STDERR $name;
    
    if($tmd){
        if($frac->[0]){
            push @$return, "GO:0005886 Plasma membrane";
        }elsif($tmd <= 1.0){
            push @$return, "GO:0031224 intrinsic to membrane";
        }elsif($tmd >  1.0){
            push @$return, "GO:0016021 integral to membrane";
        }
    }
    if($frac->[0]){
        push @$return, "GO:0005737 cytoplasm";
    }
    if(($name eq 'T') || ($frac->[1] > 70 && $name ~~ ['D/T', 'T/TD'])){
        push @$return, "GO:0005739 mitochondrion";
    }
    if($name eq 'D/T'){
        push @$return, "GO:0031982 vesicle";
    }
    if($name eq 'T/TD'){
        push @$return, "GO:0005783 endoplasmic reticulum";
    }
    if($name eq 'TD'){
        push @$return, "GO:0005634 nucleus";
    }
    if($name ~~ [qw'D/T/TD D/T/S T/S D/TD']){
        push @$return, "GO:0043226 organelle";
    }
    if($name ~~ [qw'D/T/TD/S D/TD/S T/TD/S TD/S D/S S']){
        push @$return, "GO:0043229 intracellular organelle";
    }
    
    return $return;
}

sub get_fasta_database {
    my $file = shift;
    my $ua = LWP::UserAgent->new;
    my ($params, $response, $output, $expected_length, $bytes_received);

    $params = {
            db => 'protein',
            id => join(",", @_)
        };

    $response = $ua->post(EPOST, $params);

    die "EPOST request failed" unless ($response->is_success);

    $output = $response->decoded_content;

    $params = { 
            db => 'protein',
            rettype => 'fasta',
            retmode => 'text',
        };

    $params->{WebEnv}    = $1 if ($output =~ /<WebEnv>(\S+)<\/WebEnv>/);
    $params->{query_key} = $1 if ($output =~ /<QueryKey>(\d+)<\/QueryKey>/);

    $response = $ua->request(POST(EFETCH, $params),
               sub {
                   my($chunk, $res) = @_;
                   $bytes_received += length($chunk);
                   unless (defined $expected_length) {
                      $expected_length = $res->content_length || 0;
                   }
                   if ($expected_length) {
                        printf STDERR "%d%% - ",
                                  100 * $bytes_received / $expected_length;
                   }
                   print STDERR "Downloading Fasta Database: $bytes_received bytes received\r";

                   print $file $chunk;
               });

    die "EFETCH request failed" unless ($response->is_success);

    seek($file, 0, 0);

    return $file;
}

sub configure {
    my $file = $0;
    
    my $const = {
        TMPRED_BIN => 'Location of the TMPred program: ',
        TMPRED_DAT => 'Location of the TMPred matrix.tab file: ',
        PSORT_BIN => 'Location of the PSORT II program: ',
        JOB_TEMPLATE => 'Directory to store temporary files: '
    };
    
    print "### DDF2GO CONFIGURATION ###\n";
    
    foreach (values %$const){
        print $_;
        chomp($_ = <>);
        print "";
    }
    
    $const->{JOB_TEMPLATE} = catfile($const->{JOB_TEMPLATE}, 'job.XXXX');
    
    open FILE, $file;
    my $lines = [<FILE>];
    close FILE;
    
    open FILE, '>', $file;
    foreach (@$lines){
        chomp;
        if(/use constant (\w*)/ && $const->{$1}){
            $_ = "use constant $1 => '$const->{$1}';"
        };
    
        print FILE $_;
    }
    close FILE;
    
}
1;

package Parse::Fasta;

use strict;
use warnings;

use Data::Dumper;
use Carp;

=head1 NAME

Parse::Fasta - Parses fasta sequence files

=head1 SYNOPSIS

    use Parse::Fasta;
    my $fasta = Parse::Fasta->new(*FILEHANDLE -or- $filename);
    
    while(my ($header, $sequence) = $fasta->next){
        #do something
    }
    
    print $fasta->next $print formated fasta entry;
    $fasta->get($i);
    my $old_index = $fasta->index($new_index);
    my $size = $fasta->size;

=head1 DESCRIPTION

This module reads fasta entries as they are needed.  

=head2 Methods

=over 4

=cut

=item * B<new(*FILEHANDLE -or- $filename)>

    Returns Parse::Fasta object on the given file.

=cut

sub new(\[$*]){
    my ($class, $file) = @_;
    my $self = {};

    unless( ref $file ){
        open(my $fh, $file) || croak "Can't open file $file";
        $self->{close_file} = 1;
        $file = $fh;
    }
    
    $self->{file} = $file;
    $self->{entries} = [];
    $self->{index} = 0;
    @{$self->{file_stat}}{ qw/dev ino mode nlink uid gid rdev size atime mtime ctime blksize blocks/ } = stat($file);

    {
#        local $/ = '>';
#        scalar <$file>;
        for(my $tell = tell $file; defined($_ = <$file>); $tell = tell $file){
            next unless m/^>/;
            push(@{$self->{entries}}, $tell);
        }
    }
    croak "Not a fasta file" unless @{$self->{entries}};
    
    bless( $self, $class );
    return $self;
}

=item * B<size>

    Returns number of fasta entries in file.

=cut

sub size(){
    my $self = shift;
    return scalar @{$self->{entries}};
}

=item * B<next>

=over 8

=item I<Scalar Context>

    Returns the next fasta entry in its entirety, from the '>' of the header to the '>' of the next header

=item I<List Context>

    Returns parsed header and sequence for the next fasta entry. Removes '>' from header and all whitespace from sequence.

=back

=cut

sub next(){
    my $self = shift;
    my $i = $self->{index}++;
    
    return unless defined wantarray;
    
    return wantarray ? $self->get($i) : scalar $self->get($i);
}

=item * B<index>

=item * B<index($i)>

    Returns current index. If index is suplied it returns the old index and sets the new index

=cut

sub index(;$){
    my ($self, $index) = @_;

    my $temp = $self->{index};

    $self->{index} = $index if( defined $index );
    return $temp-1;
}

=item * B<get($i)>

=over 8

=item I<Scalar Context>

    Returns the fasta entry at index $i in its entirety, from the '>' of the header to the '>' of the next header

=item I<List Context>

    Returns parsed header and sequence for the fasta entry at index $i. Removes '>' from header and all whitespace from sequence.

=back

=cut

sub get($){
    my ( $self, $i ) = @_;
    my $fh = $self->{file};
    my $str = undef;

    return if($i >= @{$self->{entries}});

    seek($fh, $self->{entries}->[$i], 0);
    read($fh, $str, ((($i+1 >= @{$self->{entries}})? $self->{file_stat}->{size} : $self->{entries}->[$i+1]) - $self->{entries}->[$i]));

    if(wantarray){
        my ($header, $sequence) = split("\n", $str, 2);
        die $header unless $sequence;    
        $sequence =~ s/\s//g;
        $header = substr($header, CORE::index($header, '>')+1 );    
        return ($header, $sequence);
    }else{
        return $str;
    }
}

sub DESTROY {
    my $self = shift;
    my $fh = $self->{file};
    close $fh if $self->{close_file};
}

1;

